(***********************************************************)
(**  Indy Server Demo                                     **)
(***********************************************************)
(**  Created by: Ching-Yuh Huang   seaturn99@yahoo.com.tw **)
(**  Created on: 13/02/2004                               **)
(**  Origional Indy Version: 9.00.10                      **)
(***********************************************************)
(**  licence :                                            **)
(***********************************************************)
(**  YTӷ~γ~,pݭnШӫHi
(**  iHNק,Yקεo{ bug  mail            **)
(**  wӫHQ׵{޳ND                             **)
(***********************************************************)

unit TSKFTPServerMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, IdUserAccounts, IdBaseComponent, IdComponent, IdTCPServer,
  IdFTPServer, IdFTPList, StdCtrls, IdSocketHandle, IdIntercept, IdLogBase,
  IdLogEvent;

type
  TfmTSKFTPServer = class(TForm)
    IdFTPServer: TIdFTPServer;
    idUserMgr: TIdUserManager;
    memLog: TMemo;
    btnSaveLog: TButton;
    SaveDialog: TSaveDialog;
    procedure IdFTPServerListDirectory(ASender: TIdFTPServerThread;
      const APath: string; ADirectoryListing: TIdFTPListItems);
    procedure IdFTPServerChangeDirectory(ASender: TIdFTPServerThread;
      var VDirectory: string);
    procedure FormCreate(Sender: TObject);
    procedure IdFTPServerStoreFile(ASender: TIdFTPServerThread;
      const AFileName: string; AAppend: Boolean; var VStream: TStream);
    procedure IdFTPServerRetrieveFile(ASender: TIdFTPServerThread;
      const AFileName: string; var VStream: TStream);
    procedure IdFTPServerMakeDirectory(ASender: TIdFTPServerThread;
      var VDirectory: string);
    procedure IdFTPServerDeleteFile(ASender: TIdFTPServerThread;
      const APathName: string);
    procedure IdFTPServerRemoveDirectory(ASender: TIdFTPServerThread;
      var VDirectory: string);
    procedure IdFTPServerAfterUserLogin(ASender: TIdFTPServerThread);
    procedure btnSaveLogClick(Sender: TObject);
    procedure IdFTPServerRenameFile(ASender: TIdFTPServerThread;
      const ARenameFromFile, ARenameToFile: string);
  private
    { Private declarations }
    procedure SendRFCReply(ASender: TIdFTPServerThread; const ATextCode, AText:
      string);
    procedure LogEvent(AMsg: string);
  public
    { Public declarations }
    function LoadUserListFromFile: boolean;
    procedure LoadWelcomeMsgFromFile;
    function LoadServerConfig: boolean;
    procedure SetFTPServerBinding(AIP: string; Aport: integer);
  end;

var
  fmTSKFTPServer: TfmTSKFTPServer;

implementation

{$R *.dfm}
uses XShadow,IniFiles, IdRFCReply, IdResourceStrings;

// Send RFC 959 Reply Message
procedure TfmTSKFTPServer.SendRFCReply(ASender: TIdFTPServerThread; const
  ATextCode, AText: string);
var
  RFCReply: TIdRFCReply;
begin
  try
    RFCReply := TIdRFCReply.Create(nil);
    RFCReply.NumericCode := StrToInt(ATextCode);
    RFCReply.TextCode := ATextCode;
    RFCReply.Text.Text := AText;
    ASender.Connection.WriteRFCReply(RFCReply);
  finally
    FreeAndNil(RFCReply);
  end;
end;

// ToDo : Save in Config.ini
const
  WelComeMessageFile = '\Config\Welcome.txt';

procedure TfmTSKFTPServer.LoadWelcomeMsgFromFile;
var
  slWelcomMsg: TStringList;
begin
  try
    slWelcomMsg := TStringList.Create;
    slWelcomMsg.LoadFromFile(GetCurrentDir + WelComeMessageFile);
    idFTPServer.Greeting.Text.Text := slWelcomMsg.Text;
  finally
    FreeAndNil(slWelcomMsg);
  end;
end;

// ToDo : Save in Config.ini
const
  UserListFile = '\Config\Accounts.txt';

  // q Accounts.txt ŪX UserID , UserName

function TfmTSKFTPServer.LoadUserListFromFile: boolean;
var
  i: integer;
  slUserList: TStringList;
  UserAccount: TidUserAccount;
begin
  Result := false;

  try
    slUserList := TStringList.Create;
    slUserList.LoadFromFile(GetCurrentDir + UserListFile);

    for i := 0 to slUserList.Count - 1 do
    begin
      UserAccount := idUserMgr.Accounts.Add;
      UserAccount.UserName := GetToken(slUserList.Strings[i], 1);
      UserAccount.Password := GetToken(slUserList.Strings[i], 2);
      UserAccount.RealName := GetToken(slUserList.Strings[i], 3);
    end;
  finally
    FreeAndNil(slUserList);
  end;

  if idUserMgr.Accounts.Count > 0 then
    Result := true;

end;

const
  ConfigFile = '\Config\Config.ini';

function TfmTSKFTPServer.LoadServerConfig: boolean;
var
  Port, DataPort: integer;
  ServerINI: TIniFile;
  ConfigFileName, IP: string;
begin
  Result := False;
  ConfigFileName := GetCurrentDir + ConfigFile;
  try
    if FileExists(ConfigFileName) then
    begin
      ServerINI := TIniFile.Create(ConfigFileName);

      IP := ServerINI.ReadString('HOST', 'IP', '');
      Port := ServerINI.ReadInteger('HOST', 'Port', 0);

      if IP = '' then
        raise Exception.Create('Can not Find IP Parameter!');

      if Port = 0 then
        Port := idFTPServer.DefaultPort;

      DataPort := ServerINI.ReadInteger('Server', 'DataPort', 0);

      if DataPort <> 0 then
        idFTPServer.DefaultPort := DataPort;

      // ]w Binding IP & Port
      SetFTPServerBinding(IP, Port);

      Result := true;
    end
    else
      raise Exception.CreateFmt('Can not Find Config File : %s',
        [ConfigFileName]);
  finally
    FreeAndNil(ServerINI);
  end;
end;


// ToDo : Save in <User>.ini , and Set User Home
const
  FTPHomeDir = 'C:\Temp\';

  // CXwؿUҦؿPW,
  // X OnChangeDirectory iܻPؿ\

procedure TfmTSKFTPServer.IdFTPServerListDirectory(
  ASender: TIdFTPServerThread; const APath: string;
  ADirectoryListing: TIdFTPListItems);
var
  flItem: TIdFTPListItem;
  SearchRec: TSearchRec;
  sDir: string;
begin
  ADirectoryListing.BeginUpdate;

  // קK]X FTPHomeDir ؿ
  sDir := PathWithoutSlash(FTPHomeDir) + APath;
  ReplaceString(sDir, '/', '\', false);

  if FindFirst(sDir + '*.*', faAnyFile, SearchRec) = 0 then
  begin
    repeat
      flItem := ADirectoryListing.Add;
      flItem.FileName := SearchRec.Name;

      // ToDo : QΩwqɮת覡,wؿ/ɮק@v޲z

      //  Unix Format
      flItem.UserPermissions := 'rwx';
      flItem.GroupPermissions := 'rwx';
      flItem.OwnerPermissions := 'rwx';
      flItem.OwnerName := 'SouthWind';
      flItem.GroupName := 'User';

      flItem.Size := SearchRec.Size;
      flItem.ModifiedDate := FileDateToDateTime(SearchRec.Time);

      if (SearchRec.Attr and faDirectory) = 0 then
      begin
        flItem.ItemType := ditFile;
      end
      else
      begin
        flItem.ItemType := ditDirectory;
      end;
    until FindNext(SearchRec) <> 0;
    FindClose(SearchRec);
  end;

  ADirectoryListing.EndUpdate;

end;

// CWD Command implementation

procedure TfmTSKFTPServer.IdFTPServerChangeDirectory(
  ASender: TIdFTPServerThread; var VDirectory: string);
var
  Path: string;
begin
  Path := VDirectory;

  // Filter '../' קK]X FTPHomeDir
  if copy(Path, 1, 3) = '../' then
  begin
    Path := PathWithoutSlash(FTPHomeDir) + ASender.CurrentDir;
    ReplaceString(Path, '/', '\', false);
    Path := ParentDirectory(Path);

    VDirectory := '/' + RelativePath(FTPHomeDir, Path);
    ReplaceString(VDirectory, '\', '/', false);
  end
  else
  begin
    Path := PathWithoutSlash(FTPHomeDir) + Path;
    ReplaceString(Path, '/', '\', false);
  end;

  if not DirectoryExists(Path) then
  begin
    SendRFCReply(ASender, '550', Format(RSLPDDirectoryDoesNotExist,
      [VDirectory]));
    // ٭ؿ,קK~ؿ
    VDirectory := ASender.CurrentDir;
  end;

end;

procedure TfmTSKFTPServer.FormCreate(Sender: TObject);
begin
  // qɮפŪX ID, Password
  LoadUserListFromFile;
  // Q idUserMgr ޲z User Account
  idFTPServer.UserAccounts := idUserMgr;

  // qɮפŪX Welcome Message
  LoadWelcomeMsgFromFile;

  // ]w LIST Format  Unix Type
  idFTpServer.EmulateSystem := ftpsUnix;
  idFTpServer.SystemType := 'Unix Type:L8';

  // Start Binding ...
  idFTPServer.Bindings.Clear;
  LoadServerConfig;

  idFTPServer.Active := true;
end;

procedure TfmTSKFTPServer.SetFTPServerBinding(AIP: string; APort: Integer);
var
  hSocketInfo: TIdSocketHandle;
begin
  hSocketInfo := idFTPServer.Bindings.Add;
  hSocketInfo.IP := AIP;
  hSocketInfo.Port := APort;
end;

procedure TfmTSKFTPServer.IdFTPServerStoreFile(ASender: TIdFTPServerThread;
  const AFileName: string; AAppend: Boolean; var VStream: TStream);
var
  tarFileName: string;
begin
  tarFileName := PathWithoutSlash(FTPHomeDir) + AFileName;
  ReplaceString(tarFileName, '/', '\', false);

  VStream := TFileStream.Create(tarFileName, fmCreate or fmOpenReadWrite);
  VStream.Position := 0;

  ASender.Connection.ReadStream(VStream, VStream.Size, false);

  LogEvent(format('User : %s Retrieve File "%s"', [ASender.Username,
    tarFileName]));
end;

procedure TfmTSKFTPServer.IdFTPServerRetrieveFile(
  ASender: TIdFTPServerThread; const AFileName: string;
  var VStream: TStream);
var
  tarFileName: string;
begin
  tarFileName := PathWithoutSlash(FTPHomeDir) + AFileName;
  ReplaceString(tarFileName, '/', '\', false);

  if FileExists(tarFileName) then
  begin
    VStream := TFileStream.Create(tarFileName, fmOpenRead);
    LogEvent(format('User : %s Retrieve File "%s"', [ASender.Username,
      tarFileName]));
  end
  else
    SendRFCReply(ASender, '550', format('%s: No such file or directory.',
      [AFileName]));

end;

procedure TfmTSKFTPServer.IdFTPServerMakeDirectory(
  ASender: TIdFTPServerThread; var VDirectory: string);
var
  Path: string;
begin
  Path := PathWithoutSlash(FTPHomeDir) + VDirectory;
  ReplaceString(Path, '/', '\', false);

  if not DirectoryExists(Path) then
  begin
    if CreateDir(Path) then
      LogEvent(format('User : %s make directory "%s"', [ASender.Username, Path]))
    else
      SendRFCReply(ASender, '550', format('%s: Can not make Directory.',
        [VDirectory]));
  end
  else
    SendRFCReply(ASender, '550', format('%s: Directory exists.', [VDirectory]));

end;

procedure TfmTSKFTPServer.IdFTPServerDeleteFile(
  ASender: TIdFTPServerThread; const APathName: string);
var
  tarFileName: string;
begin
  tarFileName := PathWithoutSlash(FTPHomeDir) + ASender.CurrentDir + '\' +
    APathName;
  ReplaceString(tarFileName, '/', '\', false);

  if FileExists(tarFileName) then
  begin
    if DeleteFile(tarFileName) then
      LogEvent(format('User : %s delete file "%s"', [ASender.Username,
        tarFileName]))
    else
      SendRFCReply(ASender, '550', Format('%s : Can not delete file.',
        [APathName]));
  end
  else
    SendRFCReply(ASender, '550', Format(RSFileNotFound,
      [APathName]));

end;

procedure TfmTSKFTPServer.IdFTPServerRemoveDirectory(
  ASender: TIdFTPServerThread; var VDirectory: string);
var
  Path: string;
begin
  Path := PathWithoutSlash(FTPHomeDir) + VDirectory;
  ReplaceString(Path, '/', '\', false);

  if DirectoryExists(Path) then
  begin
    RmDir(Path);
    if IOResult = 0 then
      LogEvent(format('User : %s make directory "%s"', [ASender.Username, Path]))
    else
      SendRFCReply(ASender, '550', format('%s: Can not Remove Directory',
        [VDirectory]));
  end
  else
    SendRFCReply(ASender, '550', format('%s: No such file or directory.',
      [VDirectory]));
end;

procedure TfmTSKFTPServer.IdFTPServerAfterUserLogin(
  ASender: TIdFTPServerThread);
begin
  ASender.HomeDir := FTPHomeDir;
  ASender.CurrentDir := '/';

  LogEvent(format('User : %s Login in Home Directory "%s"', [ASender.Username,
    ASender.CurrentDir]));
end;

procedure TfmTSKFTPServer.LogEvent(AMsg: string);
begin
  memLog.Lines.Add(AMsg);
end;

// Log in Memo

procedure TfmTSKFTPServer.btnSaveLogClick(Sender: TObject);
begin
  if SaveDialog.Execute then
    memLog.Lines.SaveToFile(SaveDialog.FileName);
end;

procedure TfmTSKFTPServer.IdFTPServerRenameFile(
  ASender: TIdFTPServerThread; const ARenameFromFile,
  ARenameToFile: string);
var
  OldName, NewName, Path: string;
begin
  Path := PathWithoutSlash(FTPHomeDir) + ASender.CurrentDir + '/';
  ReplaceString(Path, '/', '\', false);

  OldName := Path + ARenameFromFile;
  NewName := Path + ARenameToFile;

  if FileExists(Path + '\' + ARenameFromFile) then
    if RenameFile(OldName, NewName) then
      LogEvent(format('User : %s rename file : "%s" -> "%s"', [ASender.Username,
        OldName, NewName]))
    else
      SendRFCReply(ASender, '550',
        format('Can not rename file : %s',
        [OldName]))
  else
    SendRFCReply(ASender, '550', format(RSFileNotFound, [OldName]));
end;

end.

